'use strict';

function check (datum, size, alpha) {
  size = size === undefined
    ? datum.length
    : size;

  if (typeof size !== 'number') {
    throw new TypeError('size must be a number.')
  }

  if (size < 1) {
    throw new TypeError('size must be positive.')
  }

  if (size > datum.length) {
    throw new RangeError('size must be less than datum length.')
  }

  return [datum, size]
}


class Cache {
  constructor (cache = {}) {
    this._cache = cache;
  }

  get (i) {
    return this._cache[i]
  }

  set (i, v) {
    this._cache[i] = v;
  }
}


Cache.Readonly = class extends Cache {
  set () {}
};


class Period {
  constructor (size, {
    cache
  } = {}) {

    this._size = size;
    this._ma = undefined;
    this._sum = 0;
    this._length = 0;
    this._cache = cache || new Cache;
  }

  get value () {
    return this._ma
  }

  get length () {
    return this._length
  }

  push (value) {
    const index = this._length ++;
    const length = index + 1;
    const size = this._size;

    this._cache.set(index, value);

    return this._ma = length > size

      // 3. When after
      ? this._then(value, index)
      : length === size

        // 2. When is about to
        ? this._first(value, index)

        // 1. If the size of the collection is less than ma size
        : this._prepare(value, index)
  }
}


class Cumulative {
  constructor () {
    this._ma = 0;
    this._length = 0;
  }

  get value () {
    return this._ma
  }

  get length () {
    return this._length
  }

  push (value) {
    return this._ma = this._push(value)
  }
}


function periodsFactory (periodFactory) {
  // @param {Number=datum.length} setSize
  return (...args) => {
    const [datum, size] = check(...args);

    const mover = periodFactory(size, datum);

    return datum.reduce((prev, current) => {
      const m = prev.mover.push(current);
      if (m) {
        prev.ma.push(m);
      }

      return prev

    }, {
      mover,
      ma: []
    })
    .ma
  }
}

// MAsimple = (N1 + ... + Nn) / n

class SimpleCumulative extends Cumulative {
  constructor () {
    super();
    this._sum = 0;
  }

  push (value) {
    const sum = this._sum += value;
    return this._ma = sum / ++ this._length
  }
}


class SimplePeriod extends Period {
  _prepare (value, index) {
    this._sum += value;
  }

  _first (value, index) {
    return (this._sum += value) / this._size
  }

  _then (value, index) {
    const size = this._size;
    return (this._sum += value - this._cache.get(index - size)) / size
  }
}


function simple (datum) {
  const length = datum.length;
  let i = 0;
  let sum = 0;

  while (i < length) {
    sum += datum[i ++];
  }

  return sum / length
}


simple.Period = SimplePeriod;
simple.Cumulative = SimpleCumulative;
simple.Cache = Cache;

simple.periods = periodsFactory((size, datum) => {
  return new SimplePeriod(size, {
    cache: new Cache.Readonly(datum)
  })
});

// https://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average

// EMA(1) = P1
// EMA(n) = a * Pn + (1 - a) * EMA(n-1)

const denominator = n => n * (n + 1) / 2;


class WeightedPeriod extends Period {
  constructor (size, options) {
    super(size, options);

    this._numerator = 0;
    this._datum = [];
    this._denominator = denominator(size);
  }

  _prepare (value, index) {
    this._sum += value;
    this._datum.push(value);
  }

  _first (value, index) {
    this._sum += value;

    const datum = this._datum;
    datum.push(value);
    this._numerator = numerator(datum);
    this._datum = null;

    return this._numerator / this._denominator
  }

  _then (value, index) {
    const size = this._size;

    this._numerator += size * value - this._sum;
    this._sum += value - this._cache.get(index - size);
    return this._numerator / this._denominator
  }
}


function numerator (datum) {
  let numerator = 0;
  let i = 0;
  const n = datum.length;

  while (i < n) {
    numerator += datum[i] * ++ i;
  }

  return numerator
}


function weighted (datum) {
  return numerator(datum) / denominator(datum.length)
}


weighted.Period = WeightedPeriod;

weighted.periods = periodsFactory((size, datum) => {
  return new WeightedPeriod(size, {
    cache: new Cache.Readonly(datum)
  })
});

const one = (avg, offset, datum, size) => {
  let i = 0;
  let sum = 0;

  while (i < size) {
    sum += Math.pow(datum[offset + i++] - avg, 2);
  }

  return sum / size;
};

var index = ((datum, size) => {
  const ma = simple.periods(datum, size);

  return ma.map((avg, offset) => one(avg, offset, datum, size));
});

module.exports = index;
